@@ -42,13 +42,24 @@ pub fn cli() -> Command {
                 .help_heading(heading::PACKAGE_SELECTION),
         )
         .arg_manifest_path()
+        .arg_ignore_rust_version_with_help(
+            "Ignore `rust-version` specification in packages (unstable)",
+        )
         .after_help(color_print::cstr!(
             "Run `<cyan,bold>cargo help update</>` for more detailed information.\n"
         ))
 }
 
-pub fn exec(config: &mut Config, args: &ArgMatches) -> CliResult {
-    let ws = args.workspace(config)?;
+pub fn exec(gctx: &mut GlobalContext, args: &ArgMatches) -> CliResult {
+    if args.honor_rust_version().is_some() {
+        gctx.cli_unstable().fail_if_stable_opt_custom_z(
+            "--ignore-rust-version",
+            9930,
+            "msrv-policy",
+            gctx.cli_unstable().msrv_policy,
+        )?;
+    }
+    let ws = args.workspace(gctx)?;
 
     if args.is_present_with_zero_values("package") {
         print_available_packages(&ws)?;
@@ -76,7 +87,7 @@ pub fn exec(config: &mut Config, args: &ArgMatches) -> CliResult {
         to_update,
         dry_run: args.dry_run(),
         workspace: args.flag("workspace"),
-        config,
+        gctx,
     };
     ops::update_lockfile(&ws, &update_opts)?;
     Ok(())
