@@ -3,15 +3,21 @@ use std::fmt::{self, Debug, Formatter};
 use std::path::{Path, PathBuf};
 use std::task::Poll;
 
-use crate::core::source::MaybePackage;
-use crate::core::{Dependency, Package, PackageId, QueryKind, Source, SourceId, Summary};
+use crate::core::{Dependency, Package, PackageId, SourceId};
 use crate::ops;
-use crate::util::{internal, CargoResult, Config};
+use crate::sources::source::MaybePackage;
+use crate::sources::source::QueryKind;
+use crate::sources::source::Source;
+use crate::sources::IndexSummary;
+use crate::util::{internal, CargoResult, GlobalContext};
 use anyhow::Context as _;
 use cargo_util::paths;
 use filetime::FileTime;
+use gix::bstr::{BString, ByteVec};
+use gix::dir::entry::Status;
+use gix::index::entry::Stage;
 use ignore::gitignore::GitignoreBuilder;
-use tracing::{trace, warn};
+use tracing::{debug, trace, warn};
 use walkdir::WalkDir;
 
 /// A source represents one or multiple packages gathering from a given root
@@ -23,33 +29,33 @@ use walkdir::WalkDir;
 ///
 /// It also provides convenient methods like [`PathSource::list_files`] to
 /// list all files in a package, given its ability to walk the filesystem.
-pub struct PathSource<'cfg> {
+pub struct PathSource<'gctx> {
     /// The unique identifier of this source.
     source_id: SourceId,
     /// The root path of this source.
     path: PathBuf,
-    /// Whether this source has updated all package informations it may contain.
+    /// Whether this source has updated all package information it may contain.
     updated: bool,
     /// Packages that this sources has discovered.
     packages: Vec<Package>,
     /// Whether this source should discover nested packages recursively.
     /// See [`PathSource::new_recursive`] for more.
     recursive: bool,
-    config: &'cfg Config,
+    gctx: &'gctx GlobalContext,
 }
 
-impl<'cfg> PathSource<'cfg> {
+impl<'gctx> PathSource<'gctx> {
     /// Invoked with an absolute path to a directory that contains a `Cargo.toml`.
     ///
     /// This source will only return the package at precisely the `path`
     /// specified, and it will be an error if there's not a package at `path`.
-    pub fn new(path: &Path, source_id: SourceId, config: &'cfg Config) -> PathSource<'cfg> {
+    pub fn new(path: &Path, source_id: SourceId, gctx: &'gctx GlobalContext) -> PathSource<'gctx> {
         PathSource {
             source_id,
             path: path.to_path_buf(),
             updated: false,
             packages: Vec::new(),
-            config,
+            gctx,
             recursive: false,
         }
     }
@@ -62,10 +68,14 @@ impl<'cfg> PathSource<'cfg> {
     ///
     /// Note that this should be used with care and likely shouldn't be chosen
     /// by default!
-    pub fn new_recursive(root: &Path, id: SourceId, config: &'cfg Config) -> PathSource<'cfg> {
+    pub fn new_recursive(
+        root: &Path,
+        id: SourceId,
+        gctx: &'gctx GlobalContext,
+    ) -> PathSource<'gctx> {
         PathSource {
             recursive: true,
-            ..PathSource::new(root, id, config)
+            ..PathSource::new(root, id, gctx)
         }
     }
 
@@ -95,15 +105,15 @@ impl<'cfg> PathSource<'cfg> {
     }
 
     /// Returns the packages discovered by this source. It may walk the
-    /// filesystem if package informations haven't yet updated.
+    /// filesystem if package information haven't yet updated.
     pub fn read_packages(&self) -> CargoResult<Vec<Package>> {
         if self.updated {
             Ok(self.packages.clone())
         } else if self.recursive {
-            ops::read_packages(&self.path, self.source_id, self.config)
+            ops::read_packages(&self.path, self.source_id, self.gctx)
         } else {
             let path = self.path.join("Cargo.toml");
-            let (pkg, _) = ops::read_package(&path, self.source_id, self.config)?;
+            let pkg = ops::read_package(&path, self.source_id, self.gctx)?;
             Ok(vec![pkg])
         }
     }
@@ -132,7 +142,17 @@ impl<'cfg> PathSource<'cfg> {
         let root = pkg.root();
         let no_include_option = pkg.manifest().include().is_empty();
         let git_repo = if no_include_option {
-            self.discover_git_repo(root)?
+            if self
+                .gctx
+                .get_env("__CARGO_GITOXIDE_DISABLE_LIST_FILES")
+                .ok()
+                .as_deref()
+                == Some("1")
+            {
+                self.discover_git_repo(root)?.map(Git2OrGixRepository::Git2)
+            } else {
+                self.discover_gix_repo(root)?.map(Git2OrGixRepository::Gix)
+            }
         } else {
             None
         };
@@ -173,9 +193,8 @@ impl<'cfg> PathSource<'cfg> {
         };
 
         let filter = |path: &Path, is_dir: bool| {
-            let relative_path = match path.strip_prefix(root) {
-                Ok(p) => p,
-                Err(_) => return false,
+            let Ok(relative_path) = path.strip_prefix(root) else {
+                return false;
             };
 
             let rel = relative_path.as_os_str();
@@ -191,7 +210,10 @@ impl<'cfg> PathSource<'cfg> {
         // Attempt Git-prepopulate only if no `include` (see rust-lang/cargo#4135).
         if no_include_option {
             if let Some(repo) = git_repo {
-                return self.list_files_git(pkg, &repo, &filter);
+                return match repo {
+                    Git2OrGixRepository::Git2(repo) => self.list_files_git(pkg, &repo, &filter),
+                    Git2OrGixRepository::Gix(repo) => self.list_files_gix(pkg, &repo, &filter),
+                };
             }
         }
         self.list_files_walk(pkg, &filter)
@@ -223,11 +245,9 @@ impl<'cfg> PathSource<'cfg> {
         let repo_relative_path = match paths::strip_prefix_canonical(root, repo_root) {
             Ok(p) => p,
             Err(e) => {
-                tracing::warn!(
+                warn!(
                     "cannot determine if path `{:?}` is in git repo `{:?}`: {:?}",
-                    root,
-                    repo_root,
-                    e
+                    root, repo_root, e
                 );
                 return Ok(None);
             }
@@ -240,6 +260,51 @@ impl<'cfg> PathSource<'cfg> {
         Ok(None)
     }
 
+    /// Returns [`Some(gix::Repository)`](gix::Repository) if the discovered repository
+    /// (searched upwards from `root`) contains a tracked `<root>/Cargo.toml`.
+    /// Otherwise, the caller should fall back on full file list.
+    fn discover_gix_repo(&self, root: &Path) -> CargoResult<Option<gix::Repository>> {
+        let repo = match gix::ThreadSafeRepository::discover(root) {
+            Ok(repo) => repo.to_thread_local(),
+            Err(e) => {
+                tracing::debug!(
+                    "could not discover git repo at or above {}: {}",
+                    root.display(),
+                    e
+                );
+                return Ok(None);
+            }
+        };
+        let index = repo
+            .index_or_empty()
+            .with_context(|| format!("failed to open git index at {}", repo.path().display()))?;
+        let repo_root = repo.work_dir().ok_or_else(|| {
+            anyhow::format_err!(
+                "did not expect repo at {} to be bare",
+                repo.path().display()
+            )
+        })?;
+        let repo_relative_path = match paths::strip_prefix_canonical(root, repo_root) {
+            Ok(p) => p,
+            Err(e) => {
+                warn!(
+                    "cannot determine if path `{:?}` is in git repo `{:?}`: {:?}",
+                    root, repo_root, e
+                );
+                return Ok(None);
+            }
+        };
+        let manifest_path = gix::path::join_bstr_unix_pathsep(
+            gix::path::to_unix_separators_on_windows(gix::path::into_bstr(repo_relative_path)),
+            "Cargo.toml",
+        );
+        if index.entry_index_by_path(&manifest_path).is_ok() {
+            return Ok(Some(repo));
+        }
+        // Package Cargo.toml is not in git, don't use git to guide our selection.
+        Ok(None)
+    }
+
     /// Lists files relevant to building this package inside this source by
     /// consulting both Git index (tracked) or status (untracked) under
     /// a given Git repository.
@@ -251,7 +316,7 @@ impl<'cfg> PathSource<'cfg> {
         repo: &git2::Repository,
         filter: &dyn Fn(&Path, bool) -> bool,
     ) -> CargoResult<Vec<PathBuf>> {
-        warn!("list_files_git {}", pkg.package_id());
+        debug!("list_files_git {}", pkg.package_id());
         let index = repo.index()?;
         let root = repo
             .workdir()
@@ -326,7 +391,12 @@ impl<'cfg> PathSource<'cfg> {
 
             match file_path.file_name().and_then(|s| s.to_str()) {
                 // The `target` directory is never included.
-                Some("target") => continue,
+                Some("target") => {
+                    // Only filter out target if its in the package root.
+                    if file_path.parent().unwrap() == pkg_path {
+                        continue;
+                    }
+                }
 
                 // Keep track of all sub-packages found and also strip out all
                 // matches we've found so far. Note, though, that if we find
@@ -334,7 +404,7 @@ impl<'cfg> PathSource<'cfg> {
                 Some("Cargo.toml") => {
                     let path = file_path.parent().unwrap();
                     if path != pkg_path {
-                        warn!("subpackage found: {}", path.display());
+                        debug!("subpackage found: {}", path.display());
                         ret.retain(|p| !p.starts_with(path));
                         subpackages_found.push(path.to_path_buf());
                         continue;
@@ -354,15 +424,8 @@ impl<'cfg> PathSource<'cfg> {
             // symlink points to a directory.
             let is_dir = is_dir.unwrap_or_else(|| file_path.is_dir());
             if is_dir {
-                warn!("  found submodule {}", file_path.display());
-                let rel = file_path.strip_prefix(root)?;
-                let rel = rel.to_str().ok_or_else(|| {
-                    anyhow::format_err!("invalid utf-8 filename: {}", rel.display())
-                })?;
-                // Git submodules are currently only named through `/` path
-                // separators, explicitly not `\` which windows uses. Who knew?
-                let rel = rel.replace(r"\", "/");
-                match repo.find_submodule(&rel).and_then(|s| s.open()) {
+                trace!("  found directory {}", file_path.display());
+                match git2::Repository::open(&file_path) {
                     Ok(repo) => {
                         let files = self.list_files_git(pkg, &repo, filter)?;
                         ret.extend(files.into_iter());
@@ -374,7 +437,7 @@ impl<'cfg> PathSource<'cfg> {
             } else if filter(&file_path, is_dir) {
                 assert!(!is_dir);
                 // We found a file!
-                warn!("  found {}", file_path.display());
+                trace!("  found {}", file_path.display());
                 ret.push(file_path);
             }
         }
@@ -400,6 +463,143 @@ impl<'cfg> PathSource<'cfg> {
         }
     }
 
+    /// Lists files relevant to building this package inside this source by
+    /// traversing the git working tree, while avoiding ignored files.
+    ///
+    /// This looks into Git sub-repositories as well, resolving them to individual files.
+    /// Symlinks to directories will also be resolved, but walked as repositories if they
+    /// point to one to avoid picking up `.git` directories.
+    fn list_files_gix(
+        &self,
+        pkg: &Package,
+        repo: &gix::Repository,
+        filter: &dyn Fn(&Path, bool) -> bool,
+    ) -> CargoResult<Vec<PathBuf>> {
+        debug!("list_files_gix {}", pkg.package_id());
+        let options = repo
+            .dirwalk_options()?
+            .emit_untracked(gix::dir::walk::EmissionMode::Matching)
+            .emit_ignored(None)
+            .emit_tracked(true)
+            .recurse_repositories(false)
+            .symlinks_to_directories_are_ignored_like_directories(true)
+            .emit_empty_directories(false);
+        let index = repo.index_or_empty()?;
+        let root = repo
+            .work_dir()
+            .ok_or_else(|| anyhow::format_err!("can't list files on a bare repository"))?;
+        assert!(
+            root.is_absolute(),
+            "BUG: paths used internally are absolute, and the repo inherits that"
+        );
+
+        let pkg_path = pkg.root();
+        let repo_relative_pkg_path = pkg_path.strip_prefix(root).unwrap_or(Path::new(""));
+        let target_prefix = gix::path::to_unix_separators_on_windows(gix::path::into_bstr(
+            repo_relative_pkg_path.join("target/"),
+        ));
+        let package_prefix =
+            gix::path::to_unix_separators_on_windows(gix::path::into_bstr(repo_relative_pkg_path));
+
+        let pathspec = {
+            // Include the package root.
+            let mut include = BString::from(":/");
+            include.push_str(package_prefix.as_ref());
+
+            // Exclude the target directory.
+            let mut exclude = BString::from(":!/");
+            exclude.push_str(target_prefix.as_ref());
+
+            vec![include, exclude]
+        };
+
+        let mut files = Vec::<PathBuf>::new();
+        let mut subpackages_found = Vec::new();
+        for item in repo
+            .dirwalk_iter(index.clone(), pathspec, Default::default(), options)?
+            .filter(|res| {
+                // Don't include Cargo.lock if it is untracked. Packaging will
+                // generate a new one as needed.
+                res.as_ref().map_or(true, |item| {
+                    !(item.entry.status == Status::Untracked
+                        && item.entry.rela_path == "Cargo.lock")
+                })
+            })
+            .map(|res| res.map(|item| (item.entry.rela_path, item.entry.disk_kind)))
+            .chain(
+                // Append entries that might be tracked in `<pkg_root>/target/`.
+                index
+                    .prefixed_entries(target_prefix.as_ref())
+                    .unwrap_or_default()
+                    .iter()
+                    .filter(|entry| {
+                        // probably not needed as conflicts prevent this to run, but let's be explicit.
+                        entry.stage() == Stage::Unconflicted
+                    })
+                    .map(|entry| {
+                        (
+                            entry.path(&index).to_owned(),
+                            // Do not trust what's recorded in the index, enforce checking the disk.
+                            // This traversal is not part of a `status()`, and tracking things in `target/`
+                            // is rare.
+                            None,
+                        )
+                    })
+                    .map(Ok),
+            )
+        {
+            let (rela_path, kind) = item?;
+            let file_path = root.join(gix::path::from_bstr(rela_path));
+            if file_path.file_name().and_then(|name| name.to_str()) == Some("Cargo.toml") {
+                // Keep track of all sub-packages found and also strip out all
+                // matches we've found so far. Note, though, that if we find
+                // our own `Cargo.toml`, we keep going.
+                let path = file_path.parent().unwrap();
+                if path != pkg_path {
+                    debug!("subpackage found: {}", path.display());
+                    files.retain(|p| !p.starts_with(path));
+                    subpackages_found.push(path.to_path_buf());
+                    continue;
+                }
+            }
+
+            // If this file is part of any other sub-package we've found so far,
+            // skip it.
+            if subpackages_found.iter().any(|p| file_path.starts_with(p)) {
+                continue;
+            }
+
+            let is_dir = kind.map_or(false, |kind| {
+                if kind == gix::dir::entry::Kind::Symlink {
+                    // Symlinks must be checked to see if they point to a directory
+                    // we should traverse.
+                    file_path.is_dir()
+                } else {
+                    kind.is_dir()
+                }
+            });
+            if is_dir {
+                // This could be a submodule, or a sub-repository. In any case, we prefer to walk
+                // it with git-support to leverage ignored files and to avoid pulling in entire
+                // .git repositories.
+                match gix::open(&file_path) {
+                    Ok(sub_repo) => {
+                        files.extend(self.list_files_gix(pkg, &sub_repo, filter)?);
+                    }
+                    Err(_) => {
+                        self.walk(&file_path, &mut files, false, filter)?;
+                    }
+                }
+            } else if (filter)(&file_path, is_dir) {
+                assert!(!is_dir);
+                trace!("  found {}", file_path.display());
+                files.push(file_path);
+            }
+        }
+
+        return Ok(files);
+    }
+
     /// Lists files relevant to building this package inside this source by
     /// walking the filesystem from the package root path.
     ///
@@ -462,7 +662,7 @@ impl<'cfg> PathSource<'cfg> {
                     }
                 }
                 Err(err) if err.loop_ancestor().is_some() => {
-                    self.config.shell().warn(err)?;
+                    self.gctx.shell().warn(err)?;
                 }
                 Err(err) => match err.path() {
                     // If an error occurs with a path, filter it again.
@@ -530,27 +730,33 @@ impl<'cfg> PathSource<'cfg> {
     }
 }
 
-impl<'cfg> Debug for PathSource<'cfg> {
+enum Git2OrGixRepository {
+    Git2(git2::Repository),
+    Gix(gix::Repository),
+}
+
+impl<'gctx> Debug for PathSource<'gctx> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         write!(f, "the paths source")
     }
 }
 
-impl<'cfg> Source for PathSource<'cfg> {
+impl<'gctx> Source for PathSource<'gctx> {
     fn query(
         &mut self,
         dep: &Dependency,
         kind: QueryKind,
-        f: &mut dyn FnMut(Summary),
+        f: &mut dyn FnMut(IndexSummary),
     ) -> Poll<CargoResult<()>> {
         self.update()?;
         for s in self.packages.iter().map(|p| p.summary()) {
             let matched = match kind {
                 QueryKind::Exact => dep.matches(s),
-                QueryKind::Fuzzy => true,
+                QueryKind::Alternatives => true,
+                QueryKind::Normalized => dep.matches(s),
             };
             if matched {
-                f(s.clone())
+                f(IndexSummary::Candidate(s.clone()))
             }
         }
         Poll::Ready(Ok(()))
