@@ -1,9 +1,10 @@
 use cargo_test_support::compare::assert_ui;
+use cargo_test_support::current_dir;
+use cargo_test_support::file;
 use cargo_test_support::prelude::*;
+use cargo_test_support::str;
 use cargo_test_support::Project;
 
-use cargo_test_support::curr_dir;
-
 #[cargo_test]
 fn case() {
     cargo_test_support::registry::init();
@@ -14,7 +15,7 @@ fn case() {
         .rust_version("1.72")
         .publish();
 
-    let project = Project::from_template(curr_dir!().join("in"));
+    let project = Project::from_template(current_dir!().join("in"));
     let project_root = project.root();
     let cwd = &project_root;
 
@@ -24,11 +25,15 @@ fn case() {
         .arg("--ignore-rust-version")
         .arg_line("rust-version-user")
         .current_dir(cwd)
+        .env(
+            "CARGO_RESOLVER_SOMETHING_LIKE_PRECEDENCE",
+            "something-like-rust-version",
+        )
         .masquerade_as_nightly_cargo(&["msrv-policy"])
         .assert()
-        .code(101)
-        .stdout_matches_path(curr_dir!().join("stdout.log"))
-        .stderr_matches_path(curr_dir!().join("stderr.log"));
+        .code(0)
+        .stdout_matches(str![""])
+        .stderr_matches(file!["stderr.term.svg"]);
 
-    assert_ui().subset_matches(curr_dir!().join("out"), &project_root);
+    assert_ui().subset_matches(current_dir!().join("out"), &project_root);
 }
