@@ -1,7 +1,7 @@
+use cargo_test_support::file;
 use cargo_test_support::paths;
 use cargo_test_support::prelude::*;
-
-use cargo_test_support::curr_dir;
+use cargo_test_support::str;
 
 #[cargo_test]
 fn case() {
@@ -12,8 +12,8 @@ fn case() {
         .current_dir(&cwd)
         .assert()
         .code(101)
-        .stdout_matches_path(curr_dir!().join("stdout.log"))
-        .stderr_matches_path(curr_dir!().join("stderr.log"));
+        .stdout_matches(str![""])
+        .stderr_matches(file!["stderr.term.svg"]);
 
     assert!(!cwd.join("Cargo.toml").is_file());
 }
