@@ -35,9 +35,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] echo v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/echo[EXE]`
 ",
         )
@@ -59,9 +59,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] echo v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/echo[EXE]`
 ",
         )
@@ -84,7 +84,7 @@ args: []
         .with_stderr(
             "\
 [COMPILING] foo v0.0.1 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `target/debug/foo[EXE]`
 ",
         )
@@ -108,6 +108,8 @@ error: no such command: `echo`
 <tab>Did you mean `bench`?
 
 <tab>View all installed commands with `cargo --list`
+<tab>Find a package to install `echo` with `cargo search cargo-echo`
+<tab>To run the file `echo`, provide a relative path like `./echo`
 ",
         )
         .run();
@@ -136,9 +138,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] echo v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/echo[EXE]`
 ",
         )
@@ -162,7 +164,7 @@ fn warn_when_plugin_masks_manifest_on_stable() {
         .with_stdout("")
         .with_stderr(
             "\
-warning: external subcommand `echo.rs` has the appearance of a manfiest-command
+warning: external subcommand `echo.rs` has the appearance of a manifest-command
 This was previously accepted but will be phased out when `-Zscript` is stabilized.
 For more information, see issue #12207 <https://github.com/rust-lang/cargo/issues/12207>.
 ",
@@ -181,7 +183,7 @@ fn requires_nightly() {
         .with_stdout("")
         .with_stderr(
             "\
-error: running `echo.rs` requires `-Zscript`
+[ERROR] running the file `echo.rs` requires `-Zscript`
 ",
         )
         .run();
@@ -199,7 +201,7 @@ fn requires_z_flag() {
         .with_stdout("")
         .with_stderr(
             "\
-error: running `echo.rs` requires `-Zscript`
+[ERROR] running the file `echo.rs` requires `-Zscript`
 ",
         )
         .run();
@@ -208,11 +210,10 @@ error: running `echo.rs` requires `-Zscript`
 #[cargo_test]
 fn clean_output_with_edition() {
     let script = r#"#!/usr/bin/env cargo
-
-//! ```cargo
-//! [package]
-//! edition = "2018"
-//! ```
+```cargo
+[package]
+edition = "2018"
+```
 
 fn main() {
     println!("Hello world!");
@@ -230,7 +231,7 @@ fn main() {
         .with_stderr(
             "\
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -240,10 +241,9 @@ fn main() {
 #[cargo_test]
 fn warning_without_edition() {
     let script = r#"#!/usr/bin/env cargo
-
-//! ```cargo
-//! [package]
-//! ```
+```cargo
+[package]
+```
 
 fn main() {
     println!("Hello world!");
@@ -260,9 +260,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -289,9 +289,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -306,8 +306,8 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -323,9 +323,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -342,7 +342,7 @@ fn use_script_config() {
 
     let p = cargo_test_support::project()
         .file(
-            ".cargo/config",
+            ".cargo/config.toml",
             r#"
 [build]
 rustc = "non-existent-rustc"
@@ -435,9 +435,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -460,9 +460,9 @@ args: ["-NotAnArg"]
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] -NotAnArg`
 ",
         )
@@ -485,9 +485,9 @@ args: ["-NotAnArg"]
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] -NotAnArg`
 ",
         )
@@ -510,9 +510,9 @@ args: ["--help"]
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] --help`
 ",
         )
@@ -534,42 +534,177 @@ args: []
 "#,
         )
         .with_stderr(
-            r#"[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+            r#"[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] s-h-w-c- v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/s-h-w-c-[EXE]`
 "#,
         )
         .run();
 }
 
+#[cargo_test]
+fn test_name_has_leading_number() {
+    let script = ECHO_SCRIPT;
+    let p = cargo_test_support::project()
+        .file("42answer.rs", script)
+        .build();
+
+    p.cargo("-Zscript -v 42answer.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_stdout(
+            r#"bin: [..]/debug/answer[EXE]
+args: []
+"#,
+        )
+        .with_stderr(
+            r#"[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[COMPILING] answer v0.0.0 ([ROOT]/foo)
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
+[RUNNING] `[..]/debug/answer[EXE]`
+"#,
+        )
+        .run();
+}
+
+#[cargo_test]
+fn test_name_is_number() {
+    let script = ECHO_SCRIPT;
+    let p = cargo_test_support::project().file("42.rs", script).build();
+
+    p.cargo("-Zscript -v 42.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_stdout(
+            r#"bin: [..]/debug/package[EXE]
+args: []
+"#,
+        )
+        .with_stderr(
+            r#"[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[COMPILING] package v0.0.0 ([ROOT]/foo)
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
+[RUNNING] `[..]/debug/package[EXE]`
+"#,
+        )
+        .run();
+}
+
 #[cargo_test]
 fn script_like_dir() {
     let p = cargo_test_support::project()
-        .file("script.rs/foo", "something")
+        .file("foo.rs/foo", "something")
         .build();
 
-    p.cargo("-Zscript -v script.rs")
+    p.cargo("-Zscript -v foo.rs")
         .masquerade_as_nightly_cargo(&["script"])
         .with_status(101)
         .with_stderr(
             "\
-error: manifest path `script.rs` is a directory but expected a file
+[ERROR] no such file or subcommand `foo.rs`
+<tab>`foo.rs` is a directory
 ",
         )
         .run();
 }
 
 #[cargo_test]
-fn missing_script_rs() {
+fn non_existent_rs() {
     let p = cargo_test_support::project().build();
 
-    p.cargo("-Zscript -v script.rs")
+    p.cargo("-Zscript -v foo.rs")
         .masquerade_as_nightly_cargo(&["script"])
         .with_status(101)
         .with_stderr(
             "\
-[ERROR] manifest path `script.rs` does not exist
+[ERROR] no such file or subcommand `foo.rs`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn non_existent_rs_stable() {
+    let p = cargo_test_support::project().build();
+
+    p.cargo("-v foo.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stdout("")
+        .with_stderr(
+            "\
+[ERROR] no such subcommand `foo.rs`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn did_you_mean_file() {
+    let p = cargo_test_support::project()
+        .file("food.rs", ECHO_SCRIPT)
+        .build();
+
+    p.cargo("-Zscript -v foo.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stdout("")
+        .with_stderr(
+            "\
+[ERROR] no such file or subcommand `foo.rs`
+<tab>Did you mean the file `./food.rs`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn did_you_mean_file_stable() {
+    let p = cargo_test_support::project()
+        .file("food.rs", ECHO_SCRIPT)
+        .build();
+
+    p.cargo("-v foo.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stdout("")
+        .with_stderr(
+            "\
+[ERROR] no such subcommand `foo.rs`
+<tab>Did you mean the file `./food.rs` with `-Zscript`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn did_you_mean_command() {
+    let p = cargo_test_support::project().build();
+
+    p.cargo("-Zscript -v build--manifest-path=./Cargo.toml")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stdout("")
+        .with_stderr(
+            "\
+[ERROR] no such file or subcommand `build--manifest-path=./Cargo.toml`
+<tab>Did you mean the command `build --manifest-path=./Cargo.toml`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn did_you_mean_command_stable() {
+    let p = cargo_test_support::project().build();
+
+    p.cargo("-v build--manifest-path=./Cargo.toml")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stdout("")
+        .with_stderr(
+            "\
+[ERROR] no such subcommand `build--manifest-path=./Cargo.toml`
+<tab>Did you mean the command `build --manifest-path=./Cargo.toml`
 ",
         )
         .run();
@@ -579,11 +714,10 @@ fn missing_script_rs() {
 fn test_name_same_as_dependency() {
     Package::new("script", "1.0.0").publish();
     let script = r#"#!/usr/bin/env cargo
-
-//! ```cargo
-//! [dependencies]
-//! script = "1.0.0"
-//! ```
+```cargo
+[dependencies]
+script = "1.0.0"
+```
 
 fn main() {
     println!("Hello world!");
@@ -600,13 +734,14 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [UPDATING] `dummy-registry` index
+[LOCKING] 2 packages to latest compatible versions
 [DOWNLOADING] crates ...
 [DOWNLOADED] script v1.0.0 (registry `dummy-registry`)
 [COMPILING] script v1.0.0
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] --help`
 ",
         )
@@ -616,11 +751,10 @@ fn main() {
 #[cargo_test]
 fn test_path_dep() {
     let script = r#"#!/usr/bin/env cargo
-
-//! ```cargo
-//! [dependencies]
-//! bar.path = "./bar"
-//! ```
+```cargo
+[dependencies]
+bar.path = "./bar"
+```
 
 fn main() {
     println!("Hello world!");
@@ -640,10 +774,11 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[LOCKING] 2 packages to latest compatible versions
 [COMPILING] bar v0.0.1 ([ROOT]/foo/bar)
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] --help`
 ",
         )
@@ -670,9 +805,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] --help`
 ",
         )
@@ -699,9 +834,9 @@ fn main() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE] --help`
 ",
         )
@@ -724,9 +859,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[ROOT]/home/.cargo/target/[..]/debug/script[EXE]`
 ",
         )
@@ -752,9 +887,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[ROOT]/home/.cargo/target/[..]/debug/script[EXE]`
 ",
         )
@@ -815,9 +950,9 @@ fn cmd_check_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [CHECKING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 ",
         )
         .run();
@@ -876,9 +1011,9 @@ fn cmd_build_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 ",
         )
         .run();
@@ -904,9 +1039,9 @@ test result: ok. 1 passed; 0 failed; 0 ignored; 0 measured; 0 filtered out; fini
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] test [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `test` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] unittests script.rs ([..])
 ",
         )
@@ -933,7 +1068,8 @@ fn cmd_clean_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[REMOVED] [..] files, [..] total
 ",
         )
         .run();
@@ -954,7 +1090,7 @@ fn cmd_generate_lockfile_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 ",
         )
         .run();
@@ -980,7 +1116,7 @@ fn cmd_metadata_with_embedded() {
                 "default_run": null,
                 "name": "script",
                 "version": "0.0.0",
-                "id": "script[..]",
+                "id": "path+file:[..]foo#script@0.0.0",
                 "keywords": [],
                 "source": null,
                 "dependencies": [],
@@ -1018,18 +1154,18 @@ fn cmd_metadata_with_embedded() {
                 "publish": []
             }
         ],
-        "workspace_members": ["script 0.0.0 (path+file:[..]foo)"],
-        "workspace_default_members": ["script 0.0.0 (path+file:[..]foo)"],
+        "workspace_members": ["path+file:[..]foo#script@0.0.0"],
+        "workspace_default_members": ["path+file:[..]foo#script@0.0.0"],
         "resolve": {
             "nodes": [
                 {
                     "dependencies": [],
                     "deps": [],
                     "features": [],
-                    "id": "script 0.0.0 (path+file:[..]foo)"
+                    "id": "path+file:[..]foo#script@0.0.0"
                 }
             ],
-            "root": "script 0.0.0 (path+file:[..]foo)"
+            "root": "path+file:[..]foo#script@0.0.0"
         },
         "target_directory": "[ROOT]/home/.cargo/target/[..]",
         "version": 1,
@@ -1039,7 +1175,7 @@ fn cmd_metadata_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 ",
         )
         .run();
@@ -1068,7 +1204,7 @@ fn cmd_read_manifest_with_embedded() {
     "repository": null,
     "rust_version": null,
     "version":"0.0.0",
-    "id":"script[..]0.0.0[..](path+file://[..]/foo)",
+    "id":"path+file://[..]/foo#script@0.0.0",
     "keywords": [],
     "license": null,
     "license_file": null,
@@ -1095,7 +1231,7 @@ fn cmd_read_manifest_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 ",
         )
         .run();
@@ -1116,9 +1252,9 @@ args: []
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 [COMPILING] script v0.0.0 ([ROOT]/foo)
-[FINISHED] dev [unoptimized + debuginfo] target(s) in [..]s
+[FINISHED] `dev` profile [unoptimized + debuginfo] target(s) in [..]s
 [RUNNING] `[..]/debug/script[EXE]`
 ",
         )
@@ -1140,7 +1276,7 @@ script v0.0.0 ([ROOT]/foo)
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 ",
         )
         .run();
@@ -1160,7 +1296,7 @@ fn cmd_update_with_embedded() {
         )
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
 ",
         )
         .run();
@@ -1177,7 +1313,61 @@ fn cmd_verify_project_with_embedded() {
         .with_json(r#"{"success":"true"}"#)
         .with_stderr(
             "\
-[WARNING] `package.edition` is unspecifiead, defaulting to `2021`
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn cmd_pkgid_with_embedded() {
+    let p = cargo_test_support::project()
+        .file("script.rs", ECHO_SCRIPT)
+        .build();
+
+    p.cargo("-Zscript pkgid --manifest-path script.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stderr(
+            "\
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[ERROR] [ROOT]/foo/script.rs is unsupported by `cargo pkgid`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn cmd_package_with_embedded() {
+    let p = cargo_test_support::project()
+        .file("script.rs", ECHO_SCRIPT)
+        .build();
+
+    p.cargo("-Zscript package --manifest-path script.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stderr(
+            "\
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[ERROR] [ROOT]/foo/script.rs is unsupported by `cargo package`
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn cmd_publish_with_embedded() {
+    let p = cargo_test_support::project()
+        .file("script.rs", ECHO_SCRIPT)
+        .build();
+
+    p.cargo("-Zscript publish --manifest-path script.rs")
+        .masquerade_as_nightly_cargo(&["script"])
+        .with_status(101)
+        .with_stderr(
+            "\
+[WARNING] `package.edition` is unspecified, defaulting to `2021`
+[ERROR] [ROOT]/foo/script.rs is unsupported by `cargo publish`
 ",
         )
         .run();
