@@ -110,6 +110,7 @@ fn foo_bar_project(url: &str) -> Project {
                 [package]
                 name = "foo"
                 version = "0.1.0"
+                edition = "2015"
 
                 [dependencies]
                 bar = {{ git = "{url}" }}
@@ -202,7 +203,12 @@ fn known_host_works() {
     fs::write(agent.ssh_dir.join("known_hosts"), key).unwrap();
     p.cargo("fetch")
         .env("SSH_AUTH_SOCK", &agent.sock)
-        .with_stderr("[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`")
+        .with_stderr(
+            "\
+[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`
+[LOCKING] 2 packages to latest compatible versions
+",
+        )
         .run();
 }
 
@@ -274,7 +280,12 @@ fn known_host_without_port() {
     let p = foo_bar_project(&url);
     p.cargo("fetch")
         .env("SSH_AUTH_SOCK", &agent.sock)
-        .with_stderr("[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`")
+        .with_stderr(
+            "\
+[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`
+[LOCKING] 2 packages to latest compatible versions
+",
+        )
         .run();
 }
 
@@ -307,7 +318,10 @@ fn hostname_case_insensitive() {
     p.cargo("fetch")
         .env("SSH_AUTH_SOCK", &agent.sock)
         .with_stderr(&format!(
-            "[UPDATING] git repository `ssh://testuser@{hostname}:{port}/repos/bar.git`"
+            "\
+[UPDATING] git repository `ssh://testuser@{hostname}:{port}/repos/bar.git`
+[LOCKING] 2 packages to latest compatible versions
+"
         ))
         .run();
 }
@@ -381,7 +395,12 @@ Caused by:
     drop(f);
     p.cargo("fetch")
         .env("SSH_AUTH_SOCK", &agent.sock)
-        .with_stderr("[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`")
+        .with_stderr(
+            "\
+[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`
+[LOCKING] 2 packages to latest compatible versions
+",
+        )
         .run();
 }
 
@@ -406,6 +425,7 @@ fn invalid_github_key() {
                 [package]
                 name = "foo"
                 version = "0.1.0"
+                edition = "2015"
 
                 [dependencies]
                 bitflags = { git = "ssh://git@github.com/rust-lang/bitflags.git", tag = "1.3.2" }
@@ -444,6 +464,7 @@ fn bundled_github_works() {
                 [package]
                 name = "foo"
                 version = "0.1.0"
+                edition = "2015"
 
                 [dependencies]
                 bitflags = { git = "ssh://git@github.com/rust-lang/bitflags.git", tag = "1.3.2" }
@@ -552,6 +573,7 @@ Caused by:
             [package]
             name = "foo"
             version = "0.1.0"
+            edition = "2015"
 
             [dependencies]
             bitflags = { git = "ssh://git@github.com:22/rust-lang/bitflags.git", tag = "1.3.2" }
@@ -586,6 +608,11 @@ fn ssh_key_in_config() {
     );
     p.cargo("fetch")
         .env("SSH_AUTH_SOCK", &agent.sock)
-        .with_stderr("[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`")
+        .with_stderr(
+            "\
+[UPDATING] git repository `ssh://testuser@127.0.0.1:[..]/repos/bar.git`
+[LOCKING] 2 packages to latest compatible versions
+",
+        )
         .run();
 }
