@@ -1,4 +1,5 @@
 //! Read Rust code on stdin, print syntax tree on stdout.
+use ide::Edition;
 use syntax::{AstNode, SourceFile};
 
 use crate::cli::{flags, read_stdin};
@@ -7,7 +8,7 @@ impl flags::Parse {
     pub fn run(self) -> anyhow::Result<()> {
         let _p = tracing::span!(tracing::Level::INFO, "parsing").entered();
         let text = read_stdin()?;
-        let file = SourceFile::parse(&text).tree();
+        let file = SourceFile::parse(&text, Edition::CURRENT).tree();
         if !self.no_dump {
             println!("{:#?}", file.syntax());
         }
